import upygame as pygame
import urandom as random
import umachine

import res
import maps

tile_w = 8
tile_h = 8
tiles = {
    "c": pygame.surface.Surface(tile_w, tile_h, res.pixels_crate),
    "w": pygame.surface.Surface(tile_w, tile_h, res.pixels_wall),
    ".": pygame.surface.Surface(tile_w, tile_h, res.pixels_floor),
    "e": pygame.surface.Surface(tile_w, tile_h, res.pixels_exit),
    "s": pygame.surface.Surface(tile_w, tile_h, res.pixels_start),
    "b": pygame.surface.Surface(tile_w, tile_h, res.pixels_btn_off),
    "B": pygame.surface.Surface(tile_w, tile_h, res.pixels_btn_on),
    "o": pygame.surface.Surface(tile_w, tile_h, res.pixels_btndoor_closed),
    "O": pygame.surface.Surface(tile_w, tile_h, res.pixels_btndoor_open),
    "d": pygame.surface.Surface(tile_w, tile_h, res.pixels_door_closed),
    "D": pygame.surface.Surface(tile_w, tile_h, res.pixels_door_open),
    "k": pygame.surface.Surface(tile_w, tile_h, res.pixels_key),
    "n": pygame.surface.Surface(tile_w, tile_h, res.pixels_cat),
    
    # tileset 2
    "W": pygame.surface.Surface(tile_w, tile_h, res.pixels_wall2),
    "C": pygame.surface.Surface(tile_w, tile_h, res.pixels_crate2),
}

tiles["E"] = tiles["s"]
tiles["S"] = tiles["e"]

class TileSprite(object):
    def __init__(self, images, row=0, column=0):
        self.cur_image = images[0]
        self.images = images
        self.set_coord(row, column)
        self.vx = 0
        self.vy = 0
        self.keep_moving = False
        self.activated = False
        self.reached_tile = True

    def set_coord(self, row, column):
        self.row = row
        self.column = column
        self.x = column * tile_w
        self.y = row * tile_h
        self.reached_tile = True

    def update(self):
        self.reached_tile = False
        if self.vx:
           self.x += self.vx
           if self.x % tile_w == 0:
              self.column = self.x // tile_w
              if self.keep_moving:
                 self.vx = 0
                 self.keep_moving = False
                 self.reached_tile = True

        if self.vy:
           self.y += self.vy
           
           if self.y % tile_h == 0:
               self.row = self.y // tile_h
               if self.keep_moving:
                  self.vy = 0
                  self.keep_moving = False
                  self.reached_tile = True

    def move(self, vx=0, vy=0):
        self.vx = vx
        self.vy = vy
        self.keep_moving = True

    def stop(self):
        self.keep_moving = False # stop on next tile coord

class Key(TileSprite): 
    def __init__(self, row=0, column=0):
        TileSprite.__init__(self, (tiles["k"],), row, column)

class Cat(TileSprite): 
    def __init__(self, row=0, column=0):
        TileSprite.__init__(self, (tiles["n"],), row, column)

class Crate(TileSprite):
    def __init__(self, row=0, column=0, tilechar="c"):
        TileSprite.__init__(self, (tiles[tilechar],), row, column)

class Button(TileSprite):
    def __init__(self, row=0, column=0):
        TileSprite.__init__(self, (tiles["b"],tiles["B"],), row, column)
    
    def activate(self):
        self.cur_image = self.images[1]
        self.activated = True

class ButtonDoor(TileSprite):
    def __init__(self, row=0, column=0):
        TileSprite.__init__(self, (tiles["o"],tiles["O"]), row, column)
    
    def activate(self):
        self.cur_image = self.images[1]
        self.activated = True

class Door(TileSprite):
    def __init__(self, row=0, column=0):
        TileSprite.__init__(self, (tiles["d"],tiles["D"]), row, column)
    
    def activate(self):
        self.cur_image = self.images[1]
        self.activated = True
        # yuck. duplicate code...

player_images = (
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_down),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_up),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_left),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_right),
    
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_down2),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_up2),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_left2),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_right2),
    
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_down3),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_up3),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_left3),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_guy_right3),
)

evilguy_images = (
    pygame.surface.Surface(tile_w, tile_h, res.pixels_evilguy_down),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_evilguy_up),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_evilguy_left),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_evilguy_right),
    
    pygame.surface.Surface(tile_w, tile_h, res.pixels_evilguy_down2),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_evilguy_up2),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_evilguy_left2),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_evilguy_right2),
    
    pygame.surface.Surface(tile_w, tile_h, res.pixels_evilguy_down3),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_evilguy_up3),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_evilguy_left3),
    pygame.surface.Surface(tile_w, tile_h, res.pixels_evilguy_right3),
)

johnny_images = (
    pygame.surface.Surface(tile_w, tile_h, res.pixels_johnny_down),
    evilguy_images[1],
    evilguy_images[2],
    evilguy_images[3],
    
    pygame.surface.Surface(tile_w, tile_h, res.pixels_johnny_down2),
    evilguy_images[5],
    evilguy_images[6],
    evilguy_images[7],
    
    pygame.surface.Surface(tile_w, tile_h, res.pixels_johnny_down3),
    evilguy_images[9],
    evilguy_images[10],
    evilguy_images[11],
)

class Player(TileSprite):
    def __init__(self, row=0, column=0, images=player_images):
        TileSprite.__init__(self, images, row, column)
        self.keys = 0
        self.cats = 0
        self.pushing_crate = None
        self.anim_timer = 0
        self.cur_image_idx = 0

    def move(self, vx=0, vy=0):
        TileSprite.move(self, vx, vy)

        if self.vx > 0:
            self.cur_image_idx = 3
            #self.cur_image = self.images[3]
        elif self.vx < 0:
            self.cur_image_idx = 2
            #self.cur_image = self.images[2]
        elif self.vy > 0:
            self.cur_image_idx = 0
            #self.cur_image = self.images[0]
        elif self.vy < 0:
            self.cur_image_idx = 1
            #self.cur_image = self.images[1]
        elif self.pushing_crate:
            self.pushing_crate.stop()
            self.pushing_crate = None

        #self.cur_image = self.images[self.cur_image_idx]

    def update(self):
        TileSprite.update(self)
        if not self.keep_moving and self.anim_timer in (0, 20, 10):
            self.cur_image = self.images[self.cur_image_idx]
            return
        
        self.anim_timer += 1
        if self.anim_timer > 15:
            self.cur_image = self.images[self.cur_image_idx + 8]
        elif self.anim_timer > 10:
            self.cur_image = self.images[self.cur_image_idx]
        elif self.anim_timer > 5:
            self.cur_image = self.images[self.cur_image_idx + 4]
        else:
            self.cur_image = self.images[self.cur_image_idx]
        
        if self.anim_timer > 20:
            self.anim_timer = 0


class Ai(Player):
    def __init__(self, row=0, column=0, is_johnny=True):
        if is_johnny:
            Player.__init__(self, row, column, johnny_images)
        else:
            Player.__init__(self, row, column, evilguy_images)
        self.is_johnny = is_johnny
    
    def get_turn_checks(self):
        # johnny tries to turn left
        # left, forward, backward, right
        if self.cur_image_idx == 0:
            turn = (1, 0) # facing down -> move to right side of screen
            forward = (0, 1)
        elif self.cur_image_idx == 1:
            turn = (-1, 0) # facing up -> move to left side of screen
            forward = (0, -1)
        elif self.cur_image_idx == 2:
            turn = (0, 1) # facing left -> move to bottom of screen
            forward = (-1, 0)
        else: #if ai.cur_image_idx == 3:
            turn = (0, -1) # facing right -> move to top of screen
            forward = (1, 0)
        turnback = (-turn[0], -turn[1])
        
        # evilguy tries to turn right
        if not self.is_johnny:
            turnback = turn
            turn = (-turn[0], -turn[1])
        
        backward = (-forward[0], -forward[1])
        
        return (turn, forward, turnback, backward)

class Map(object):
    blocking = " wsSW"

    def __init__(self, w, h, map_data):
        self.map_data = map_data
        self.width = w
        self.height = h
        self.offset_x = (110 - self.width * tile_w) // 2
        self.offset_y = (88 - self.height * tile_h) // 2

    def get_tile_pos(self, tile, start=0):
        idx = self.map_data.index(tile, start)
        row = idx // self.width 
        col = idx % self.width
        return row, col

    def draw(self, surface):
        # bottom layer
        for row in range(self.height):
            for col in range(self.width):
                tile = self.map_data[(row * self.width) + col]
                if tile in "cCknt":
                    tile = "."
                if tile in "S":
                    surface.blit(tiles[tile], col*tile_w + self.offset_x, row*tile_h + self.offset_y, 0, False, True)
                if tile in ".wesEW":
                    surface.blit(tiles[tile], col*tile_w + self.offset_x, row*tile_h + self.offset_y)
                
    def is_blocking(self, col, row):
        idx = (row * self.width) + col
        return self.map_data[idx] in self.blocking
        #return False


class FadeEffect:
    """Black rectangle that grows/shrinks to fade out / fade in screen
    """
    
    x = 0
    y = 0
    w = 110
    h = 0
    max_h = 88
    max_w = 110
    dir = 0
    color = 0
    
    def setup(self, dir):
        if dir > 0:
            self.h = 0 #fadein
        else:
            self.h = self.max_h #fadeout
        self.dir = dir

    def update(self):
        self.h += self.dir
        return self.h > 0 and self.h <= self.max_h
    
    def draw(self, surface):
        surface.fill(self.color, pygame.Rect(self.x, self.y, self.w, self.h))
            
            

FADE_EFFECT = FadeEffect()